@@ -23,6 +23,8 @@
 #include <stdio.h>
 #include <ctype.h>
 
+#define USE_XHTML(opt) (opt->flags & HTML_USE_XHTML)
+
 struct html_renderopt {
 	void *extra;
 
@@ -32,7 +34,6 @@ struct html_renderopt {
 	} toc_data;
 
 	unsigned int flags;
-	const char *close_tag;
 };
 
 static inline void
@@ -223,9 +224,10 @@ rndr_blockcode_github(struct buf *ob, struct buf *text, struct buf *lang, void *
 static void
 rndr_blockquote(struct buf *ob, struct buf *text, void *opaque)
 {
+	if (ob->size) bufputc(ob, '\n');
 	BUFPUTSL(ob, "<blockquote>\n");
 	if (text) bufput(ob, text->data, text->size);
-	BUFPUTSL(ob, "</blockquote>");
+	BUFPUTSL(ob, "</blockquote>\n");
 }
 
 static int
@@ -272,6 +274,14 @@ rndr_emphasis(struct buf *ob, struct buf *text, void *opaque)
 	return 1;
 }
 
+static int
+rndr_linebreak(struct buf *ob, void *opaque)
+{
+	struct html_renderopt *options = opaque;	
+	bufputs(ob, USE_XHTML(options) ? "<br/>\n" : "<br>\n");
+	return 1;
+}
+
 static void
 rndr_header(struct buf *ob, struct buf *text, int level, void *opaque)
 {
@@ -355,11 +365,14 @@ rndr_paragraph(struct buf *ob, struct buf *text, void *opaque)
 			if (i > org)
 				bufput(ob, text->data + org, i - org);
 
-			if (i >= text->size)
+			/*
+			 * do not insert a line break if this newline
+			 * is the last character on the paragraph
+			 */
+			if (i >= text->size - 1)
 				break;
-
-			BUFPUTSL(ob, "<br");
-			bufputs(ob, options->close_tag);
+			
+			rndr_linebreak(ob, opaque);
 			i++;
 		}
 	} else {
@@ -398,8 +411,7 @@ rndr_hrule(struct buf *ob, void *opaque)
 {
 	struct html_renderopt *options = opaque;	
 	if (ob->size) bufputc(ob, '\n');
-	BUFPUTSL(ob, "<hr");
-	bufputs(ob, options->close_tag);
+	bufputs(ob, USE_XHTML(options) ? "<hr/>\n" : "<hr>\n");
 }
 
 static int
@@ -416,17 +428,7 @@ rndr_image(struct buf *ob, struct buf *link, struct buf *title, struct buf *alt,
 		BUFPUTSL(ob, "\" title=\"");
 		sdhtml_escape(ob, title->data, title->size); }
 
-	bufputc(ob, '"');
-	bufputs(ob, options->close_tag);
-	return 1;
-}
-
-static int
-rndr_linebreak(struct buf *ob, void *opaque)
-{
-	struct html_renderopt *options = opaque;	
-	BUFPUTSL(ob, "<br");
-	bufputs(ob, options->close_tag);
+	bufputs(ob, USE_XHTML(options) ? "\"/>" : "\">");
 	return 1;
 }
 
@@ -458,26 +460,24 @@ rndr_table(struct buf *ob, struct buf *header, struct buf *body, void *opaque)
 	BUFPUTSL(ob, "<table><thead>\n");
 	if (header)
 		bufput(ob, header->data, header->size);
-	BUFPUTSL(ob, "\n</thead><tbody>\n");
+	BUFPUTSL(ob, "</thead><tbody>\n");
 	if (body)
 		bufput(ob, body->data, body->size);
-	BUFPUTSL(ob, "\n</tbody></table>");
+	BUFPUTSL(ob, "</tbody></table>\n");
 }
 
 static void
 rndr_tablerow(struct buf *ob, struct buf *text, void *opaque)
 {
-	if (ob->size) bufputc(ob, '\n');
 	BUFPUTSL(ob, "<tr>\n");
 	if (text)
 		bufput(ob, text->data, text->size);
-	BUFPUTSL(ob, "\n</tr>");
+	BUFPUTSL(ob, "</tr>\n");
 }
 
 static void
 rndr_tablecell(struct buf *ob, struct buf *text, int align, void *opaque)
 {
-	if (ob->size) bufputc(ob, '\n');
 	switch (align) {
 	case MKD_TABLE_ALIGN_L:
 		BUFPUTSL(ob, "<td align=\"left\">");
@@ -498,7 +498,7 @@ rndr_tablecell(struct buf *ob, struct buf *text, int align, void *opaque)
 
 	if (text)
 		bufput(ob, text->data, text->size);
-	BUFPUTSL(ob, "</td>");
+	BUFPUTSL(ob, "</td>\n");
 }
 
 static int
@@ -606,9 +606,6 @@ sdhtml_toc_renderer(struct mkd_renderer *renderer, void *extra)
 void
 sdhtml_renderer(struct mkd_renderer *renderer, unsigned int render_flags, void *extra)
 {
-	static const char *xhtml_close = "/>\n";
-	static const char *html_close = ">\n";
-
 	static const struct mkd_renderer renderer_default = {
 		rndr_blockcode,
 		rndr_blockquote,
@@ -646,7 +643,6 @@ sdhtml_renderer(struct mkd_renderer *renderer, unsigned int render_flags, void *
 	struct html_renderopt *options;	
 	options = calloc(1, sizeof(struct html_renderopt));
 	options->flags = render_flags;
-	options->close_tag = (render_flags & HTML_USE_XHTML) ? xhtml_close : html_close;
 	options->extra = extra;
 
 	memcpy(renderer, &renderer_default, sizeof(struct mkd_renderer));
